using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Linq;

using MVR;
using MeshVR;

using GPUTools.Cloth.Scripts;

namespace Stopper {
    public static class ClothingScriptExtensions {

        public static T ReplaceDazSkinWrap<T>(this MVRScript script, DAZSkinWrap originalSkinWrap) where T : SkinWrapReplacer {
            // Reloading the plugin gives us a new class, so we have to match by name
            foreach(DAZSkinWrap existingReplacer in originalSkinWrap.GetComponentsInChildren<DAZSkinWrap>(true)) {
                if($"{existingReplacer.GetType()}" == $"{typeof(T)}") {
                    UnityEngine.Object.Destroy(existingReplacer);
                }
            }

            T wrapReplacer = originalSkinWrap.gameObject.AddComponent<T>();
            wrapReplacer.enabled = true;
            wrapReplacer.replacedSkinWrap = originalSkinWrap;

            SwapSkinWrap(script, originalSkinWrap, wrapReplacer);

            return wrapReplacer;
        }

        public static void InitSkinWrapSwitcher(this MVRScript script, DAZSkinWrap skinWrap) {
            var wrapSwitcher = skinWrap.GetComponent<DAZSkinWrapSwitcher>(); 

            foreach(var otherSkinWrap in skinWrap.GetComponents<DAZSkinWrap>()) {
                if(otherSkinWrap != skinWrap) {
                    otherSkinWrap.enabled = false;
                }
            }
            skinWrap.enabled = true;
            skinWrap.draw = true;
            if(wrapSwitcher != null) {
                wrapSwitcher.Init();
            }
        }

        public static void SwapSkinWrap(this MVRScript script, DAZSkinWrap originalSkinWrap, DAZSkinWrap newSkinWrap) {
            var wrapSwitcher = originalSkinWrap.GetComponent<DAZSkinWrapSwitcher>(); 
            originalSkinWrap.enabled = false;
            newSkinWrap.enabled = true;
            if(wrapSwitcher != null) {
                
                wrapSwitcher.SetCurrentWrapName(newSkinWrap.wrapName);
            } else {
                foreach (DAZSkinWrapMaterialOptions dazskinWrapMaterialOptions in originalSkinWrap.GetComponents<DAZSkinWrapMaterialOptions>())
                {
                    dazskinWrapMaterialOptions.skinWrap = newSkinWrap;
                }
                DAZSkinWrapControl skinWrapControl = originalSkinWrap.GetComponent<DAZSkinWrapControl>();
                if (skinWrapControl != null)
                {
                    skinWrapControl.wrap = newSkinWrap;
                }
                foreach (ClothSettings clothSettings in originalSkinWrap.GetComponents<ClothSettings>())
                {
                    clothSettings.MeshProvider.PreCalcProvider = newSkinWrap;
                }
            }

            newSkinWrap.enabled = true;
            newSkinWrap.draw = true;

            originalSkinWrap.enabled = false;
            originalSkinWrap.draw = false;
            
            newSkinWrap.FlushBuffers();
        }
    }
}