using System;
using UnityEngine;
using UnityEngine.Experimental.Rendering;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using SimpleJSON;
using System.Linq;
using MVR;
using MeshVR;
using MVR.FileManagementSecure;
using GPUTools.Cloth.Scripts;
using Unity.Collections;
using System.Diagnostics;
using Leap.Unity;
using UnityEngine.Experimental.PlayerLoop;

namespace Stopper {
    public class SkinWrapReplacer : DAZSkinWrap {

        private DAZSkinWrap _replacedSkinWrap;

        public DAZSkinWrap replacedSkinWrap {
            get {
                return _replacedSkinWrap;
            } set {
                if(_replacedSkinWrap != value) {
                    _replacedSkinWrap = value;
                    CopyAttributes(value);
                }
            }
        }
                
        public void CopyAttributes(DAZSkinWrap other) {
            if(other == null) {
                return;
            }
            this.wrapStore = other.wrapStore;
            this.wrapName = other.wrapName;
            this.draw = other.draw;
            this.debug = other.debug;
            this.debugStartVert = other.debugStartVert;
            this.debugStopVert = other.debugStopVert;
            this.debugTan1 = other.debugTan1;
            this.debugTan2 = other.debugTan2;
            this.debugSize = other.debugSize;
            this.dazMesh = other.dazMesh;
            this.skinTransform = other.skinTransform;
            this.skin = other.skin;
            this.forceRawSkinVerts = other.forceRawSkinVerts;
            this.defaultAdditionalThicknessMultiplier = other.defaultAdditionalThicknessMultiplier;
            this.additionalThicknessMultiplier = other.additionalThicknessMultiplier;
            this.defaultSurfaceOffset = other.defaultSurfaceOffset;
            this.surfaceOffset = other.surfaceOffset;
            this.smoothOuterLoops = other.smoothOuterLoops;
            this.laplacianSmoothPasses = other.laplacianSmoothPasses;
            this.laplacianSmoothBeta = other.laplacianSmoothBeta;
            this.springSmoothPasses = other.springSmoothPasses;
            this.springSmoothFactor = other.springSmoothFactor;
            this.useSpring2 = other.useSpring2;
            this.spring2SmoothFactor = other.spring2SmoothFactor;
            this.moveToSurface = other.moveToSurface;
            this.moveToSurfaceOffset = other.moveToSurfaceOffset;
            this.autoColliderToEnable = other.autoColliderToEnable;
            this.autoColliderToDisable = other.autoColliderToDisable;
            this.forceCPU = other.forceCPU;
            
            this.morphCopyFromTransform = other.morphCopyFromTransform;
            this.morphCopyFrom = other.morphCopyFrom;
            this.vertexIndexLimit = other.vertexIndexLimit;
            this.wrapCheckNormals = other.wrapCheckNormals;
            this.wrapToSkinnedVertices = other.wrapToSkinnedVertices;
            this.wrapToMorphedVertices = other.wrapToMorphedVertices;
            this.wrapToDisabledMaterials = other.wrapToDisabledMaterials;
            this.assetSavePath = other.assetSavePath;


            this.GPUSkinWrapper = other.GPUSkinWrapper;
            this.GPUMeshCompute = other.GPUMeshCompute;

            this.GPUuseSimpleMaterial = other.GPUuseSimpleMaterial;
            this.GPUsimpleMaterial = other.GPUsimpleMaterial;
            this.GPUmaterials = other.GPUmaterials;
            this.simTextures = other.simTextures;
            this.GPUAutoSwapShader = other.GPUAutoSwapShader;
            this.onlyUpdateEnabledMaterials = other.onlyUpdateEnabledMaterials;
            this.materialsEnabled = other.materialsEnabled;
            this.recalculateNormals = other.recalculateNormals;
            this.recalculateTangents = other.recalculateTangents;
            this._numMaterials = this.GPUmaterials.Length;
        }

        // Accessors
        public ComputeBuffer drawVerticesBuffer {
            get {
                return this._drawVerticesBuffer;
            }
        }

        public ComputeBuffer verticesBuffer1 {
            get {
                return this._verticesBuffer1;
            }
        }

        public ComputeBuffer surfaceNormalsBuffer {
            get {
                return this._surfaceNormalsBuffer;
            }
        }

        public void UpdateWrapBuffer() {
            if(this.wrapStore != null) {
                if (this._wrapVerticesBuffer != null) {
                    this._wrapVerticesBuffer.SetData(this.wrapStore.wrapVertices);
                }
                this.startingWrapStore = this.wrapStore;
            }
        }

        protected void OnDestroy() {
            // Give consumers a chance to copy needed buffers before they go away
            if(preDestroyHandlers != null) {
                preDestroyHandlers();
            }
            base.OnDestroy();
        }

        public PreDestroyHandlers preDestroyHandlers;
    
        public delegate void PreDestroyHandlers();


        // Taken from decompiled VAM code

        protected void LateUpdate(){
            // If this is a child, we rely on the parent to call this
           

            if ((this.draw || this.debug) && this.skin != null && !this._updateDrawDisabled && this.dazMesh != null && !this.isWrapping && this.wrapStore != null && this.wrapStore.wrapVertices != null)
            {
                if (this.autoColliderToEnable != null)
                {
                    this.autoColliderToEnable.on = true;
                }
                if (this.autoColliderToDisable != null)
                {
                    this.autoColliderToDisable.on = false;
                }
                if (this.skin.skinMethod == DAZSkinV2.SkinMethod.CPU || this.forceCPU || !Application.isPlaying)
                {
                    this.UpdateVerts(false);
                    if (this.draw)
                    {
                        this.DrawMesh();
                    }
                }
                else
                {
                    if (this.skin.delayDisplayOneFrame && this.GPUSkinWrapper != null && this._delayedVertsBuffer != null)
                    {
                        this.GPUSkinWrapper.SetBuffer(this._copyKernel, "inVerts", this._drawVerticesBuffer);
                        this.GPUSkinWrapper.SetBuffer(this._copyKernel, "outVerts", this._delayedVertsBuffer);
                        this.GPUSkinWrapper.Dispatch(this._copyKernel, this.numVertThreadGroups, 1, 1);
                        if (this._normalsBuffer != null)
                        {
                            this.GPUSkinWrapper.SetBuffer(this._copyKernel, "inVerts", this._normalsBuffer);
                            this.GPUSkinWrapper.SetBuffer(this._copyKernel, "outVerts", this._delayedNormalsBuffer);
                            this.GPUSkinWrapper.Dispatch(this._copyKernel, this.numVertThreadGroups, 1, 1);
                        }
                        if (this._tangentsBuffer != null)
                        {
                            this.GPUMeshCompute.SetBuffer(this._copyTangentsKernel, "inTangents", this._tangentsBuffer);
                            this.GPUMeshCompute.SetBuffer(this._copyTangentsKernel, "outTangents", this._delayedTangentsBuffer);
                            this.GPUMeshCompute.Dispatch(this._copyTangentsKernel, this.numVertThreadGroups, 1, 1);
                        }
                    }
                    this.UpdateVertsGPU(true);
                    if (this.draw)
                    {
                        this.DrawMeshGPU(false);
                    }
                }
            }

        }

        public override void PostProcessDispatch(ComputeBuffer finalVerts)
        {
            if (this.drawInPostProcess && this._wrapVerticesBuffer != null)
            {
                if (!this.provideToWorldMatrices)
                {
                    if (this.smoothOuterLoops > 0 && (this.laplacianSmoothPasses > 0 || this.springSmoothPasses > 0))
                    {
                        this.DoSmoothingGPU(finalVerts);
                    }
                    else
                    {
                        this._drawVerticesBuffer = finalVerts;
                    }
                    this.mapVerticesGPU.Map(this._drawVerticesBuffer);
                    this.RunNormalTangentRecalc();
                }
                this._updateDrawDisabled = true;
                this.DrawMeshGPU(true);
            }
        }

        protected void DrawMeshGPU(bool noDelay = false)
        {
            if (!this._renderSuspend)
            {
                if (this.materialVertsBuffers == null)
                {
                    this.materialVertsBuffers = new Dictionary<int, ComputeBuffer>();
                }
                if (this.materialNormalsBuffers == null)
                {
                    this.materialNormalsBuffers = new Dictionary<int, ComputeBuffer>();
                }
                if (this.materialTangentsBuffers == null)
                {
                    this.materialTangentsBuffers = new Dictionary<int, ComputeBuffer>();
                }
                if (DAZSkinWrap.staticDraw)
                {
                    Matrix4x4 identity = Matrix4x4.identity;
                    this.dazMesh.DrawMorphedUVMappedMesh(identity);
                }
                else if (this.mesh != null)
                {
                    Matrix4x4 identity2 = Matrix4x4.identity;
                    if (this.dazMesh.drawFromBone != null)
                    {
                        Bounds bounds = this.mesh.bounds;
                        bounds.center = this.dazMesh.drawFromBone.transform.position;
                        Vector3 size = bounds.size;
                        size.x = 0.1f;
                        size.y = 0.1f;
                        size.z = 0.1f;
                        bounds.size = size;
                        this.mesh.bounds = bounds;
                    }
                    for (int i = 0; i < this.mesh.subMeshCount; i++)
                    {
                        if (this.GPUuseSimpleMaterial && this.GPUsimpleMaterial)
                        {
                            if (this.skin.delayDisplayOneFrame && !noDelay)
                            {
                                this.GPUsimpleMaterial.SetBuffer("verts", this._delayedVertsBuffer);
                                this.GPUsimpleMaterial.SetBuffer("normals", this._delayedNormalsBuffer);
                                this.GPUsimpleMaterial.SetBuffer("tangents", this._delayedTangentsBuffer);
                            }
                            else
                            {
                                this.GPUsimpleMaterial.SetBuffer("verts", this._drawVerticesBuffer);
                                this.GPUsimpleMaterial.SetBuffer("normals", this._normalsBuffer);
                                this.GPUsimpleMaterial.SetBuffer("tangents", this._tangentsBuffer);
                            }
                            Graphics.DrawMesh(this.mesh, identity2, this.GPUsimpleMaterial, 0, null, i, null, this.dazMesh.castShadows, this.dazMesh.receiveShadows);
                        }
                        else if (this.materialsEnabled != null && this.materialsEnabled[i] && this.GPUmaterials[i] != null)
                        {
                            if (this.skin != null && this.skin.delayDisplayOneFrame && !noDelay)
                            {
                                ComputeBuffer computeBuffer;
                                this.materialVertsBuffers.TryGetValue(i, out computeBuffer);
                                if (computeBuffer != this._delayedVertsBuffer)
                                {
                                    this.GPUmaterials[i].SetBuffer("verts", this._delayedVertsBuffer);
                                    this.materialVertsBuffers.Remove(i);
                                    this.materialVertsBuffers.Add(i, this._delayedVertsBuffer);
                                }
                                this.materialNormalsBuffers.TryGetValue(i, out computeBuffer);
                                if (computeBuffer != this._delayedNormalsBuffer)
                                {
                                    this.GPUmaterials[i].SetBuffer("normals", this._delayedNormalsBuffer);
                                    this.materialNormalsBuffers.Remove(i);
                                    this.materialNormalsBuffers.Add(i, this._delayedNormalsBuffer);
                                }
                                this.materialTangentsBuffers.TryGetValue(i, out computeBuffer);
                                if (computeBuffer != this._delayedTangentsBuffer)
                                {
                                    this.GPUmaterials[i].SetBuffer("tangents", this._delayedTangentsBuffer);
                                    this.materialTangentsBuffers.Remove(i);
                                    this.materialTangentsBuffers.Add(i, this._delayedTangentsBuffer);
                                }
                            }
                            else
                            {
                                ComputeBuffer computeBuffer2;
                                this.materialVertsBuffers.TryGetValue(i, out computeBuffer2);
                                if (computeBuffer2 != this._drawVerticesBuffer)
                                {
                                    this.GPUmaterials[i].SetBuffer("verts", this._drawVerticesBuffer);
                                    this.materialVertsBuffers.Remove(i);
                                    this.materialVertsBuffers.Add(i, this._drawVerticesBuffer);
                                }
                                this.materialNormalsBuffers.TryGetValue(i, out computeBuffer2);
                                if (computeBuffer2 != this._normalsBuffer)
                                {
                                    this.GPUmaterials[i].SetBuffer("normals", this._normalsBuffer);
                                    this.materialNormalsBuffers.Remove(i);
                                    this.materialNormalsBuffers.Add(i, this._normalsBuffer);
                                }
                                this.materialTangentsBuffers.TryGetValue(i, out computeBuffer2);
                                if (computeBuffer2 != this._tangentsBuffer)
                                {
                                    this.GPUmaterials[i].SetBuffer("tangents", this._tangentsBuffer);
                                    this.materialTangentsBuffers.Remove(i);
                                    this.materialTangentsBuffers.Add(i, this._tangentsBuffer);
                                }
                            }
                            Graphics.DrawMesh(this.mesh, identity2, this.GPUmaterials[i], this.gameObject.layer, null, i, null, this.dazMesh.castShadows, this.dazMesh.receiveShadows);
                        }
                    }
                }
            }
        }
    }
}